/*
 *  Part of ase2rv source code,
 *  Copyright (C) 2000 Alexander Kroeller (alilein@gmx.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <print.h>
#include <world.h>
#include <helpfunc.h>

#include "aseprint.h"
#include "asekeys.h"
#include "aseparser.h"
#include "asemesh.h"
#include "asematerial.h"
#include "mainconv.h"


AppOptions appopt;


ASEMesh *parseMesh(ASEParser &p)
{
  ASEMesh *mesh;

  mesh=new ASEMesh; MEM(mesh);
  mesh->parse(p);
  mesh->checkSanity();
  return mesh;
}

bool stringInIncNodes(const char *str)
{
  int i;

  for(i=0;i<appopt.num_inc_nodes;i++)
    if(strcasecmp(appopt.include_nodes[i], str)==0)
      {
	appopt.touched_node[i]=true;
	return true;
      };
  return false;
}

void parseGeomObject(ASEParser &p, World *w)
{
  ASEMesh *mymesh=NULL;
  Mesh *rvmesh;
  int mat_id=0;
  int incl_this= (appopt.num_inc_nodes > 0) ? 0 : 1; /* 0 unknown, 1 yes, 2 no */

  if(p.curtok != PT_BBEGIN)
    sxpanic("GEOMOBJECT section begins without '{'");
  p.genKeyToken(true);

  do
    {
      if(p.tok_key == AK_NODE_NAME)
	{
	  p.gentoken();
	  if(p.curtok == PT_STRING)
	    {
	      if(appopt.list_mode)
		aprintf("%s\n",p.tok_text);
	      else
		{
		  if((appopt.num_inc_nodes==0) || stringInIncNodes(p.tok_text))
		    {
		      incl_this=1;
		      aprintf(" (now inside object '%s')\n",p.tok_text);
		    }
		  else
		    {
		      incl_this=2;
		      aprintf(" (skipping object '%s')\n",p.tok_text);
		    };
		      
		};
	    }
	  else
	    sxpanic("%s is no string??", id2keyname(AK_NODE_NAME));
	  p.genKeyToken(true);
	}
      else if(p.tok_key == AK_MASTER_MAT_ID)
	{
	  p.gentoken();
	  if(p.curtok != PT_INTEGER)
	    sxpanic("%s is no integer", id2keyname(AK_MASTER_MAT_ID));
	  mat_id=p.tok_int;
	  p.genKeyToken(true);
	}
      else if((p.tok_key == AK_MESH) && (!appopt.list_mode))
	{
	  p.gentoken();
	  if(incl_this == 1)
	    {
	      if(mymesh != NULL)
		sxpanic("Multiple meshes in one GEOMOBJECT");
	      mymesh=parseMesh(p);
	    }
	  else if(incl_this==0)
	    sxpanic("Mesh before node name");
	  p.genKeyToken(false);
	}
      else
	p.genKeyToken(true);
    }
  while((p.curtok != PT_BEND)&&(p.curtok != PT_EOF));

  if(mymesh != NULL)
    {
      rvmesh=mymesh->createMesh(mat_id, appopt.wmode);
      w->addMesh(rvmesh);
      delete mymesh;
    };
  p.gentoken();
}






void ase2world(ASEParser &p, World *w)
{
  while(p.curtok != PT_EOF)
    {
      p.genKeyToken(false);

      if(p.curtok != PT_KEYWORD)
	{
	  p.error("$$");
	  sxpanic("key finding failure");};

      if(p.tok_key == AK_VERSION)
	{
	  p.gentoken();
	  if(p.curtok != PT_INTEGER)
	    sxpanic("ascii export version is no integer");
	  if(p.tok_int != 200)
	    interest("ASE Export Version is %d\n",p.tok_int);
	  p.gentoken();
	}
      else if(p.tok_key == AK_MATERIAL_LIST)
	{
	  p.gentoken();
	  parseMaterialList(p);
	}
      else if(p.tok_key == AK_GEOMOBJECT)
	{
	  p.gentoken();
	  parseGeomObject(p,w);
	}
      else
	p.gentoken();
      p.genKeyToken(false);
    };
}


void deleteAppOptions(void)
{
  int i;

  for(i=0;i<appopt.num_inc_nodes;i++)
    if(!appopt.touched_node[i])
      aprintf("Warning: Object '%s' not found\n",appopt.include_nodes[i]);
  
  delete appopt.include_nodes;
  delete appopt.touched_node;
}

World *runapp(void)
{
  ASEParser p(appopt.ase_name);
  World *w;

  if(!appopt.list_mode)
    {
      w = new World; MEM(w);
    }
  else
    w=NULL;

  ase2world(p,w);

  deleteMaterials();

  deleteAppOptions();

  return w;
}


void welcome(const char *appname)
{
  /* 80:  123456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 */
  aprintf("  /-------------------------------------------\n");
  aprintf("  | %s version 0.3, Copyright (C) 2000 Alexander Kroeller and Steven Ellis\n",appname);
  aprintf("  | %s comes with ABSOLUTELY NO WARRANTY; for details see 'doc\\GPL.html'.\n",appname);
  aprintf("  \\-------------------------------------------\n");
}


void usage(const char *appname, const char *outputtype)
{
  /* 80:  123456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 */
  aprintf("Usage: %s [-v] [-l] [-sg] [-morph|-ali] file.ase [object] [object]\n\n", appname);
  aprintf("  converts .ase files to re-volt %s files\n\n", outputtype);
  aprintf("  specify -v to make %s more verbose\n\n",appname);
  aprintf("  specify -l to make %s simply list all object and doing nothing more\n\n",appname);
  aprintf("  specify -sg to make %s dump out the smoothing group meanings\n\n",appname);
  aprintf("  If you want only some objects included in the %s file, give the names\n",outputtype);
  aprintf("  of the according objects after the .ase file name. If you want all objects\n");
  aprintf("  to be included, omit the object specification.\n\n");
  aprintf("  If you want to use surface properties and transparency,\n");
  aprintf("  you have to specify either -morph or -ali (can be abbreviated with\n");
  aprintf("  -m or -a). The documenatation lists the differences.\n");
  exit(1);
}


bool parseArgs(int argc, char **argv,
	       const char *vis_suff, const char *im_suff, const char *ncp_suff)
{
  int i;
  bool sgdump_only=false;

  msglevel=1;
  appopt.ase_name[0]=0;
  appopt.wmode = None;
  appopt.list_mode = false;

  appopt.include_nodes = new char*[argc]; MEM(appopt.include_nodes);
  appopt.touched_node  = new bool[argc]; MEM(appopt.touched_node);
  appopt.num_inc_nodes = 0;

  if(argc == 1)
    return false;

  for(i=1;i<argc;i++)
    {
      if(argv[i][0] == '-')
	{
	  if(strcasecmp(argv[i],"-v")==0)
	    msglevel=2;
	  else if(strcasecmp(argv[i],"-l")==0)
	    appopt.list_mode = true;
	  else if((strcasecmp(argv[i],"-morph")==0)||(strcasecmp(argv[i],"-m")==0))
	    appopt.wmode = Morph;
	  else if((strcasecmp(argv[i],"-ali")==0)||(strcasecmp(argv[i],"-a")==0))
	    appopt.wmode = Ali;
	  else if(strcasecmp(argv[i],"-sg")==0)
	    sgdump_only=true;
	  else if(strcasecmp(argv[i],"-ase")==0)
	    aprintf=asemode_printf;
	  else
	    {
	      printf("Error: option '%s' is unknown\n",argv[i]);
	      return false;
	    };
	}
      else
	{
	  if(appopt.ase_name[0]==0)
	    strcpy(appopt.ase_name, argv[i]);
	  else
	    {
	      appopt.include_nodes[ appopt.num_inc_nodes ]=argv[i];
	      appopt.touched_node[ appopt.num_inc_nodes ]=false;
	      appopt.num_inc_nodes++;
	    };
	};
    };

  if(sgdump_only)
    {
      dump_sg_tables();
      exit(0);
    };

  if(appopt.ase_name[0]==0)
    {
      printf("Error: Specify an input file, please!\n");
      exit(1);
    };

  if(!filehastype(appopt.ase_name,"ase"))
    {
      printf("Error: Specify an input file with .ase extension, please!\n");
      exit(1);
    };

  strcpy(appopt.vis_name, appopt.ase_name);
  breakfname(appopt.vis_name);
  strcpy(appopt.im_name, appopt.vis_name);
  strcpy(appopt.ncp_name, appopt.vis_name);
 
  strcat(appopt.vis_name, vis_suff);
  strcat(appopt.im_name, im_suff);
  strcat(appopt.ncp_name, ncp_suff);

  return true;
}
